#include "board.h"
#include "bitboard.h"
#include "move.h"
#include <iostream>
#include <string.h>
#include <stdio.h>

using namespace std;

const int JOKER_MOVE = 127;



void ProcessString(char* str);

#pragma pack(1)

struct move {
	char rank, file; // 0 if not used, else the row/file of the piece
	char to;
	char piece; // 127 for a '*' move
	char score; // +3 for !! ... -3 for ??	
	char promote;

	int operator==(move& m){ 
		return rank == m.rank && file == m.file && to == m.to && piece == m.piece && promote == m.promote && score == m.score;
	}
	int operator!=(move& m){ 
		return !((*this) == m);
	}
};

#pragma pack()


int CompareMove(move& m1, int m2);



void PrintMove(move& m){
	switch(m.piece){
	case PAWN : break;
	case ROOK : cout << 'R'; break;
	case KNIGHT : cout << 'N'; break;
	case BISHOP : cout << 'B'; break;
	case KING : cout << 'K'; break;
	case QUEEN : cout << 'Q'; break;
	case JOKER_MOVE : cout << "* "; return;
	default:
		cout << "wrong piece type "; return;
	};

	if (m.rank || m.file){
		if (m.rank) cout << (char)(m.rank + 'a');
	}

	cout << (char)(File(m.to)+'a') << (char)(Rank(m.to)+'1');

	switch (m.promote){
	case QUEEN: cout << "=Q"; break;
	case ROOK: cout << "=R"; break;
	case KNIGHT : cout << "=N"; break;
	case BISHOP: cout << "=B"; break;
	};

	switch(m.score){
	case -3 : cout << "??"; break;
	case -2 : cout << "?"; break;
	case -1 : cout << "?!"; break;
	case  0 : cout << ""; break;
	case  1 : cout << "!?"; break;
	case  2 : cout << "!"; break;
	case  3 : cout << "!!"; break;
	}

	cout << " ";
}

class MoveNode {
public:
	move m;
	MoveNode* brother;
	MoveNode* son;

	MoveNode() { brother = son = NULL;}
	~MoveNode() { 
		if (brother) delete brother; 
		if(son) delete son; 
	}

	MoveNode* AddSon(move& move); 
	MoveNode* AddBrother(move& move);

	MoveNode* FindBrother(int move);
	MoveNode* FindBrother2(int move);

	void ScoreMovesRec(int* m_list, int n, int* current_moves, int* scores, int n_moves);
	
	void Print(int n, int firstone = 1);

	int CountNodes();

	void Write(FILE* fp);
	void Read(FILE* fp);
};

move tmp_move_list[300];

static int line_counter = 0;

void MoveNode::Print(int n, int firstone){
	tmp_move_list[n] = m;

	if (son){
		son->Print(n+1);
	} else {
		cout << "[" << line_counter++ << "] ";
		for (int i = 0; i <= n; i++)
			PrintMove(tmp_move_list[i]);
		cout << endl;
	}

	if (firstone)
	for (MoveNode* p = brother; p != NULL; p = p->brother)
		p->Print(n,0);
}


MoveNode* MoveNode::AddSon(move& move){
	if (son == NULL) {
		son = new MoveNode; 
		son->m = move;
		return son;
	}

	MoveNode* p = son;
	for ( ; p->m != move && p->brother ; p = p->brother)
		;
	
	if (p->m == move) return p; // no need to add an existing move

	Assert(p->brother == NULL);

	p->brother = new MoveNode;
	p->brother->m = move;
	return p->brother;
}


MoveNode* MoveNode::AddBrother(move& move){
	MoveNode* p = this;

	if (p->m.piece == 0) { // root move not initialized
		p->m = move;
		return p;
	}

	for ( ; p->m != move && p->brother; p = p->brother)
		;

	if (p->m == move) return p; // already in there

	Assert(p->brother == NULL);

	p->brother = new MoveNode;
	p->brother->m = move;
	return p->brother;
}


MoveNode* MoveNode::FindBrother(int move){
	for (MoveNode* p = this; p; p = p->brother)
		if (CompareMove(p->m,move)) 
			return p;

	return NULL;
}

MoveNode* MoveNode::FindBrother2(int move){
	for (MoveNode* p = this; p; p = p->brother)
		if (CompareMove(p->m,move) && p->m.score != 0) 
			return p;

	return NULL;
}

int MoveNode::CountNodes(){
	int count = 1;
	if (son) count+= son->CountNodes();
	if (brother) count += brother->CountNodes();

	return count;
}


void MoveNode::Write(FILE* fp){
	static char zero = 0;
	static char one = 1;

	fwrite(&m, sizeof(m), 1, fp);

	if (brother){
		fwrite(&one, 1, 1, fp);
		brother->Write(fp);
	} else 
		fwrite(&zero, 1, 1, fp);

	if (son){
		fwrite(&one, 1, 1, fp);
		son->Write(fp);
	} else 
		fwrite(&zero, 1, 1, fp);
}

void MoveNode::Read(FILE* fp){
	char x;

	if (fread(&m, sizeof(m), 1, fp) != 1){
		cout << "Damaged file!" << endl;
		return;
	}

	fread(&x, 1, 1, fp);
	if (x){
		brother = new MoveNode();
		brother->Read(fp);
	}
	fread(&x, 1, 1, fp);
	if (x){
		son = new MoveNode();
		son->Read(fp);
	}
}



class CompiledMoveList {
public:
	MoveNode root;

	void AddMovelist(int n, move* m);
	void ScoreMoves(int* movelist, int* score, int n_movelist, int* moves_so_far, int move_number);

	void Read(char* filename);
	void Write(char* filename);
};

static CompiledMoveList clist;


void CompiledMoveList::AddMovelist(int n, move* m){
	if (n == 0) return;

	MoveNode* p = root.AddBrother(m[0]);
	for (int i = 1 ; i < n; i++){
		p = p->AddSon(m[i]);
	}
}


void MoveNode::ScoreMovesRec(int* m_list, int n, int* current_moves, int* scores, int n_moves){
	if (n == 0) {
		for (int i = 0; i < n_moves; i++)
			if (CompareMove(m, current_moves[i]))
				if (m.score){
					PrintMove(m);
					scores[i] = 10000000*m.score;
				} 

		if (brother)
			brother->ScoreMovesRec(m_list, n, current_moves, scores, n_moves);

		return;
	}

	for (MoveNode* p = this; p; p = p->brother)
		if (CompareMove(p->m, *m_list))
			if (p->son)
				p->son->ScoreMovesRec(m_list+1, n-1, current_moves, scores, n_moves);
}

void CompiledMoveList::ScoreMoves(int* movelist, int* score, int n_movelist, int* moves_so_far, int move_number){
	int i;

	cout << "Score Moves :" << endl;
	cout << "Move number = " << move_number << endl;
	cout << "Moves so far : [ "; 
	for (i = 0; i < move_number; i++) PrintMove(moves_so_far[i]);
	cout << "]" << endl;

	cout << "master book moves found in this position : [";

	root.ScoreMovesRec(moves_so_far, move_number, movelist, score, n_movelist);

	cout << "]" << endl;
}

void CompiledMoveList::Read(char* filename){
	FILE* fp = fopen(filename, "rb");

	if (!fp) { cout << "Cannot read " << filename << endl; return; }

	root.Read(fp);

	fclose(fp);
}

void CompiledMoveList::Write(char* filename){
	FILE* fp = fopen(filename, "wb");

	if (!fp) { cout << "Cannot write " << filename << endl; return; }

	root.Write(fp);

	fclose(fp);
}



int CompareMove(move& m1, int m2){
	if (m1.piece == JOKER_MOVE) return 1;

	if (m1.piece != Piece(m2)) return 0;
	if (m1.to != To(m2)) return 0;

	int f = File(From(m2));
	int r = Rank(From(m2));

	if (m1.file && m1.file != (File(From(m2))+1)) return 0;
	if (m1.rank && m1.rank != (Rank(From(m2))+1)) return 0;

	if (Promote(m2) != m1.promote) return 0;

	return 1;
}


static int str_count = 0;

static void ProcessString(char* str){
	char* str0 = str;

	static move movearray[500];
	memset(movearray, 0, sizeof(movearray));

	move* m_list = movearray;

	#define SKIP_WHITESPACE	while (*str == '	' || *str == ' ' || *str == 10 || *str == '+') str++;

	#define ERROR { cout << "Error parsing the string " << str0 << " in line " << __LINE__ << endl; return; }

	#define IsFile(c) ((c) >= 'a' && (c) <= 'h')
	#define IsRank(c) ((c) >= '1' && (c) <= '8')

	#define RankChar(c) (((c)-'1')+1)
	#define FileChar(c) (((c)-'a')+1)

	if (*str == '1') str++;
	if (*str == '.') str++;

	for (int move_number = 0 ; 1 ; m_list++, move_number++){
		SKIP_WHITESPACE

		if (str[0] == 0) break;

		m_list->file = 0;
		m_list->rank = 0;

		if (*str == '*'){
			m_list->piece = JOKER_MOVE;
			m_list->to = 0;
			m_list->score = 0;
			str++; 
			continue;
		} 

		//cout << str << endl;
	
		switch(*str){
		case 'N' : m_list->piece = KNIGHT; break;
		case 'B' : m_list->piece = BISHOP; break;
		case 'R' : m_list->piece = ROOK; break;
		case 'Q' : m_list->piece = QUEEN; break;
		case 'O' :
		case 'o' :
		case 'K' : m_list->piece = KING; break;
		default  : m_list->piece = PAWN; break;
		}

		if (*str == 'O' || *str == 'o' || *str == '0'){
			if (str[1] == '-' && (str[2] == 'O' || str[2] == 'o' || str[2] == '0')){
				if (str[3] == '-' && (str[4] == 'O' || str[4] == 'o' || str[4] == '0')){
					m_list->to = (move_number & 1) ? C8 : C1;
					str+=5;
				} else {
					m_list->to = (move_number & 1) ? G8 : G1;
					str+=3;
				}
			} else
				ERROR

		} else {
			if (m_list->piece != PAWN){
				str++;

				/* We have 2 cases like in :
						1) Nfxe4 or N3xe4 or Nfe4 or N3e4
						2) Ne4 or Nxe4
				*/

				if (str[1] == 'x'){
					if (IsRank(str[0]))
						m_list->rank = RankChar(str[0]);
					else if (IsFile(str[0]))
						m_list->file = FileChar(str[0]);
					else
						ERROR
					str++;
					str++;
				} else if ((IsFile(str[0]) || IsRank(str[0]))&&IsFile(str[1])){
					if (IsRank(str[0]))
						m_list->rank = RankChar(str[0]);
					else if (IsFile(str[0]))
						m_list->file = FileChar(str[0]);
					else
						ERROR
					str++;
				}


				if (str[0] == 'x') str++;
		
				if (!IsFile(str[0]) || !IsRank(str[1]))
					ERROR

				m_list->to = (str[0]-'a') + ((str[1]-'1')*8);

				str += 2;
			} else {
				/*	We have 3 cases like in:
						1) axb4
						2) e4
						3) e8=Q
				*/
			
				if (str[1] == 'x'){
					if (!IsFile(str[0])) ERROR
					m_list->file = FileChar(str[0]);
					str++;
					str++;
				}

				if (!IsFile(str[0])) ERROR
				if (!IsRank(str[1])) ERROR

				m_list->to = (str[0]-'a') + ((str[1]-'1')*8);

				str++;
				str++;

				if (m_list->to >= A8 || m_list->to <= H1) { // promotion
					if (*str == '=') str++;
					switch (*str){
					case 'Q':
					case 'q':
						m_list->promote = QUEEN; break;
					case 'R':
					case 'r':
						m_list->promote = ROOK; break;
					case 'B':
					case 'b':
						m_list->promote = BISHOP; break;
					case 'N':
					case 'n':
						m_list->promote = KNIGHT; break;
					default:
						ERROR
					}
					str++;
				}
			}
		}

		if (str[0] == '+') {
			str++;
		}

		if (str[0] == '!')
			if (str[1] == '!')
				m_list->score = 3;
			else if (str[1] == '?')
				m_list->score = 1;
			else 
				m_list->score = 2;
		else if (str[0] == '?')
			if (str[1] == '?')
				m_list->score = -3;
			else if (str[1] == '!')
				m_list->score = -1;
			else 
				m_list->score = -2;
		else
			m_list->score = 0;

		while (*str == '!' || *str == '?') str++;
	}


	str_count++;

	clist.AddMovelist(m_list-movearray, movearray);

	#if 0

	char* x = strstr(str0, "\xA");
	if (x) *x = 0;

	cout << "AddMoveList [" << str0 << "]" << endl;


	MoveNode* p = &clist.root;

	cout << "The move list is:" << endl;
	cout << "[----------------------]" << endl;

	clist.root.Print(0);

	cout << "[----------------------]" << endl << endl;

	#endif
}



#include <sys/stat.h>
#include <time.h>
#include <io.h>

int same_time(char* fname1, char* fname2){
	struct _stat buf1, buf2;

	if (_stat(fname1, &buf1) || _stat(fname2, &buf2))
		return 0;

	return difftime(buf2.st_mtime, buf1.st_mtime) > 0;
}


void LearnBook(char* fname, char* compiled_file){
	
	if (!same_time(fname, compiled_file)){
		cout << "(Re)building " << fname << " from " << compiled_file << endl;

		FILE* fp = fopen(fname, "rt");
		if (!fp){
			cout << "Cannot open " << fname << " ! " << endl;
			return;
		}

		char str[1024];

		while(!feof(fp)){
			if (fgets(str, 1024, fp) == NULL) 
				break;
			else if (strlen(str) < 3) 
				continue;
			else if (strstr(str, "[") || strstr(str, "@")  || strstr(str, ";"))
				;
			else 
				ProcessString(str);
		}

		fclose(fp);

		clist.Write(compiled_file);
	} else {
		cout << "Reading the master opening book from " << compiled_file << endl;

		clist.Read(compiled_file);
	}

	#if 1
	//clist.root.Print(0);

	//cout << line_counter << " lines processed !" << endl;
	//cout << "should be = " << str_count << endl;
	cout << clist.root.CountNodes() << " nodes" << endl;
	#endif

	//cout << n << " parsed lines" << endl;

	//Print();
}


void ScoreMoves(int* movelist, int* score, int n_movelist, int* moves_so_far, int move_number){
	clist.ScoreMoves(movelist, score, n_movelist, moves_so_far, move_number);
}

